<?php
/**
 * Plugin PDF_VERSION pour Spip 3.x
 * Licence GPL (c) 2016 Cedric
 *
 */

if (!defined("_ECRIRE_INC_VERSION")) return;


/**
 * Appel de PDFShift  par API sur un autre site SPIP
 * @param string $html_file
 * @param string $pdf_file
 * @param array $args
 */
function pdf_version_services_pdfshift_dist($html_file, $pdf_file, $args = array()) {

	include_spip('inc/config');
	$logfile = 'pdfv_pdfshift';

	if($pdfshift_api_key = lire_config('pdf_version/pdfshift_api_key')) {

		$pdfshift_api_url = "https://api.pdfshift.io/v3/convert/pdf";

		// Appel de pdfshift via API http

		$html = "";
		$pdfshift_options = [
			'source' => '(...)',
			'use_print' => false,
			'delay' => 500,
			'format' => $args['--page-size'] ?? 'A4',
			'landscape' => empty($args['--orientation']) ? false : ($args['--orientation'] === 'Portrait' ? false : true),
			'margin' => implode(' ', [$args['--margin-top'], $args['--margin-right'], $args['--margin-bottom'], $args['--margin-left']]),
		];

		// margin et format semblent pas vraiment pris en compte, on doublonne avec une directive css
		$pdfshift_options['css'] = "@page{size:".$pdfshift_options['format'].";margin:".$pdfshift_options['margin']."}";

		// todo : substitution des variables dans le header et footer
		/*
			Variable 	Description
			{{date}} 	Formatted print date
			{{title}} 	Title of the HTML document
			{{url}} 	Page URL
			{{page}} 	Current page
			{{total}} 	Total number of pages
		*/
		$vars = [
			'@page@' => '{{page}}',
			'@total_page@' => '{{total}}',
		];

		if (isset($args['--header-html'])) {
			lire_fichier($args['--header-html'], $html);
			$html = pdf_version_renomme_variables($vars, $html);

			// il faut reinserer les margin left/right du document dans le footer car il est par defaut sans margin
			$mleft = $args['--margin-left'];
			$mright = $args['--margin-right'];
			$style_add = "<style>#header {margin-left:calc(0.75 * {$mleft});margin-right:calc(0.75 * {$mright});}</style>";
			if ($p = strpos($html, "</head>")) {
				$html = substr_replace($html, $style_add, $p, 0);
			}
			else {
				$html .= $style_add;
			}
			$pdfshift_options['header'] = ['source' => $html, 'height' => $args['--header-spacing'] ?? '1px'];
			if (is_numeric($pdfshift_options['header']['height'])) {
				$pdfshift_options['header']['height'] .= 'px';
			}
		}
		if (isset($args['--footer-html'])) {
			lire_fichier($args['--footer-html'], $html);
			$html = pdf_version_renomme_variables($vars, $html);

			// il faut reinserer les margin left/right du document dans le footer car il est par defaut sans margin
			$mleft = $args['--margin-left'];
			$mright = $args['--margin-right'];
			$mbottom = $args['--margin-bottom'];
			$style_add = "<style>#footer {margin-left:calc(0.75 * {$mleft});margin-right:calc(0.75 * {$mright});margin-bottom: calc(0.5 * $mbottom)}</style>";
			if ($p = strpos($html, "</head>")) {
				$html = substr_replace($html, $style_add, $p, 0);
			}
			else {
				$html .= $style_add;
			}
			$pdfshift_options['footer'] = ['source' => $html, 'height' => $args['--footer-spacing'] ?? '1px'];
			if (is_numeric($pdfshift_options['footer']['height'])) {
				$pdfshift_options['footer']['height'] .= 'px';
			}
		}

		if (lire_config('pdf_version/pdfshift_sandbox')) {
			$pdfshift_options['sandbox'] = true;
		}

		spip_log($pdfshift_api_url . " $html_file -> $pdf_file : options=" . json_encode($pdfshift_options), $logfile . _LOG_DEBUG);

		// lire le html et le renseigner dans les arguments
		$html = "";
		lire_fichier($html_file, $html);
		$pdfshift_options['source'] = $html;


		// un watermark ?
		$watermarks = [
			'pdf' => [
				'--pdfshit-watermark-source'
			],
			'texte' => [
				'--pdfshit-watermark-text',
				'--pdfshit-watermark-font_size',
				'--pdfshit-watermark-font_family',
				'--pdfshit-watermark-font_color',
				'--pdfshit-watermark-font_opacity',
				'--pdfshit-watermark-font_bold',
				'--pdfshit-watermark-font_italic',
				'--pdfshit-watermark-offset_x',
				'--pdfshit-watermark-offset_y',
				'--pdfshit-watermark-rotate'
			],
			'image' => [
				'--pdfshit-watermark-image',
				'--pdfshit-watermark-offset_x',
				'--pdfshit-watermark-offset_y',
				'--pdfshit-watermark-rotate'
			],
		];
		foreach ($watermarks as $type => $vars) {
			// si on a un --pdfshit-watermark-source en base64 dans le source c'est mieux de le nettoyer
			$watermark = pdf_version_collecter_variables_head($vars, $pdfshift_options['source'], $type === 'pdf' ? true : false);
			$first_var = reset($vars);
			if (!empty($watermark[$first_var])) {
				$pdfshift_options['watermark'] = [];
				foreach ($watermark as $k => $v) {
					$k = substr($k, strlen('--pdfshit-watermark-'));
					$pdfshift_options['watermark'][$k] = $v;
				}
				break;
			}
		}

		$outline_titles = pdfshift_outline_hn($pdfshift_options['source']);
		$result = true;
		$curl = curl_init();
		curl_setopt_array($curl, array(
		    CURLOPT_URL => $pdfshift_api_url,
		    CURLOPT_RETURNTRANSFER => true,
		    CURLOPT_POST => true,
		    CURLOPT_POSTFIELDS => json_encode($pdfshift_options),
		    CURLOPT_HTTPHEADER => array('Content-Type:application/json'),
		    CURLOPT_USERPWD => "api:$pdfshift_api_key"
		));

		$response = curl_exec($curl);
		$res = curl_getinfo($curl);
		$trace = json_encode($res);

		if (!empty($res['http_code'])
			and $res['http_code'] === 200
			and $response !== false) {

			pdfshift_postraite_outline($response, $outline_titles);
			file_put_contents("{$pdf_file}.tmp", $response);
		}
		else {
			$result = false;
			// en cas d'erreur pdfshift renvoie un json avec le code d'erreur et le message, on le loge donc
			if ($response !== false and strlen($response)) {
				$trace .= " | " . $response;
			}
		}

		if ($result) {
			@rename("{$pdf_file}.tmp", $pdf_file);
		}
		else {
			@unlink("{$pdf_file}.tmp");
		}

		spip_log($pdfshift_api_url . " $html_file -> $pdf_file : $trace", $logfile . ($result ? '' : _LOG_ERREUR));

		return $result;
	}

	spip_log("service pdfshift mal configure",$logfile . _LOG_ERREUR);
	return false;
}

/**
 * Ajouter une ancre id devant chaque titre <hn> du html pour générer un outline automatique
 * post génération par pdfshift
 *
 * @param string &$html
 * @return array
 */
function pdfshift_outline_hn(&$html) {

	$html_outline = "";
	$titles = [];
	if (preg_match_all(",<h([123456])\b[^>]*>(.*?)</h\\1>,ims", $html, $matches, PREG_SET_ORDER | PREG_OFFSET_CAPTURE)) {
		foreach ($matches as $m) {
			$title = [
				'pos' => $m[0][1],
				'niveau' => $m[1][0],
				'texte' => textebrut($m[2][0]),
				'raw' => $m[0][0],
			];
			$title['anchor'] = 'tit' . $title['niveau'] . '-' . substr(md5($title['pos'] . ':' . $title['raw']), 0, 8);
			$titles[] = $title;
		}

		foreach (array_reverse($titles) as $t) {
			$anchor = $t['anchor'];
			$html = substr_replace($html, "<span id=\"$anchor\"></span>", $t['pos'], 0);
			$html_outline = "<li><a href=\"#$anchor\">".$t['texte']."</a></li>\n" . $html_outline;
		}

		// pour que les ancres restents dans le pdf il faut un lien qui pointe dessus
		// on génère donc un plan en fin de document, mais non visible
		$html_outline = "<div class='outline'"
			. " style='position: absolute;left:0;bottom:0;width: 1px;height: 1px;font-size: 1px;overflow: hidden'"
			."><ul>$html_outline</ul></div>";
		$html = str_replace("</body>", $html_outline . "</body>", $html);
	}

	#spip_log(json_encode($titles), 'dbgoutline');

	return $titles;
}

function pdfshift_postraite_outline(&$pdf_source, $titles) {
	if (empty($titles)) {
		return;
	}

	$outline_declaration = null;
	// detecter la déclaration de l'outline
	if ($p = strpos($pdf_source, '/Outlines ')) {
		$chunk = substr($pdf_source, $p, 1000);
		if (preg_match(',^/Outlines (\d+) 0 R\n,', $chunk, $m)) {
			$outline_declaration = [
				'pos' => $p,
				'root' => $m[1],
				'length' => strlen($m[0]),
				'raw' => $m[0]
			];
		}
	}

	if (!$outline_declaration) {
		return;
	}

	// Trouver le dernier numero d'objet
	$lastObj = 0;
	$posAfterLastObj = null;

	$length = strlen($pdf_source);
	$lastp = strlen($pdf_source);
	$nbobj = 0;
	while (($p = strrpos($pdf_source, "endobj", $lastp - 1 - $length)) !== false
	  and ($lastObj === 0 or $nbobj++ < 10)) {
		$chunk = substr($pdf_source, $p, $lastp - $p);
		if (preg_match(",^endobj\s+(\d+) 0 obj,", $chunk, $m)) {
			$objNumber = intval($m[1]);
			$posAfterLastObj = $p + strlen($chunk) + strlen('endobj');
			var_dump("Found obj $objNumber posAfter $posAfterLastObj");
			if ($objNumber > $lastObj) {
				$lastObj = $objNumber;
			}
		}
		$lastp = $p;
	}

	if (empty($lastObj)) {
		return;
	}

	// TODO : generer le outline
	$outlines = [];
	foreach ($titles as $title) {
		$outlines[] = [
			'l' => $title['niveau'] - 1,
			't' => $title['texte'],
			'd' => '#' . $title['anchor']
		];
	}

	$bookmarkWriter = new BookmarkWriter($lastObj);
	if ($outlineRoot = $bookmarkWriter->writeBookmarks($outlines)) {
		// modifier la declaration de l'outline
		$pdf_source = substr_replace($pdf_source, "/Outlines $outlineRoot 0 R\n/PageMode /UseOutlines\n", $outline_declaration['pos'], $outline_declaration['length']);
		// inserer l'outline
		$pdf_source = substr_replace($pdf_source, "\n" . $bookmarkWriter->getBuffer(), $posAfterLastObj, 0);
	}
}

class BookmarkWriter
{

	protected array $outlines = [];

	protected int $objectNumber = 0;

	protected string $buffer = "";

	public function __construct($lastObj)
	{
		$this->objectNumber = $lastObj;
	}

	protected function write($s, $ln = true)
	{
		$this->buffer .= $s . ($ln ? "\n" : '');
	}

	protected function writeObject()
	{
		$obj_id = ++$this->objectNumber;
		$this->write($obj_id . ' 0 obj');
	}

	public function getBuffer() {
		return $this->buffer;
	}

	public function utf16BigEndianTextString($s) // _UTF16BEtextstring
	{
		$s = $this->utf8ToUtf16BigEndian($s, true);

		return '(' . $this->escape($s) . ')';
	}

	// Converts UTF-8 strings to UTF16-BE.
	public function utf8ToUtf16BigEndian($str, $setbom = true) // UTF8ToUTF16BE
	{
		/*
		if ($this->mpdf->checkSIP && preg_match("/([\x{20000}-\x{2FFFF}])/u", $str)) {
			if (!in_array($this->mpdf->currentfontfamily, ['gb', 'big5', 'sjis', 'uhc', 'gbB', 'big5B', 'sjisB', 'uhcB', 'gbI', 'big5I', 'sjisI', 'uhcI',
				'gbBI', 'big5BI', 'sjisBI', 'uhcBI'])) {
				$str = preg_replace("/[\x{20000}-\x{2FFFF}]/u", chr(0), $str);
			}
		}
		*/
		if (/*$this->mpdf->checkSMP &&*/ preg_match("/([\x{10000}-\x{1FFFF}])/u", $str)) {
			$str = preg_replace("/[\x{10000}-\x{1FFFF}]/u", chr(0), $str);
		}

		$outstr = ''; // string to be returned
		if ($setbom) {
			$outstr .= "\xFE\xFF"; // Byte Order Mark (BOM)
		}

		$outstr .= mb_convert_encoding($str, 'UTF-16BE', 'UTF-8');

		return $outstr;
	}

	public function escape($s) // _escape
	{
		return strtr($s, [')' => '\\)', '(' => '\\(', '\\' => '\\\\', chr(13) => '\r']);
	}



	public function writeBookmarks($outlines)
	{
		$nb = count($outlines);
		if ($nb === 0) {
			return 0;
		}


		$bmo = $outlines;
		$this->outlines = [];
		$lastlevel = -1;
		for ($i = 0; $i < count($bmo); $i++) {
			if ($bmo[$i]['l'] > 0) {
				while ($bmo[$i]['l'] - $lastlevel > 1) { // If jump down more than one level, insert a new entry
					$new = $bmo[$i];
					$new['t'] = "[" . $new['t'] . "]"; // Put [] around text/title to highlight
					$new['l'] = $lastlevel + 1;
					$lastlevel++;
					$this->outlines[] = $new;
				}
			}
			$this->outlines[] = $bmo[$i];
			$lastlevel = $bmo[$i]['l'];
		}
		$nb = count($this->outlines);

		$lru = [];
		$level = 0;
		foreach ($this->outlines as $i => $o) {
			if ($o['l'] > 0) {
				$parent = $lru[$o['l'] - 1];
				// Set parent and last pointers
				$this->outlines[$i]['parent'] = $parent;
				$this->outlines[$parent]['last'] = $i;
				if ($o['l'] > $level) {
					// Level increasing: set first pointer
					$this->outlines[$parent]['first'] = $i;
				}
			} else {
				$this->outlines[$i]['parent'] = $nb;
			}
			if ($o['l'] <= $level and $i > 0) {
				// Set prev and next pointers
				$prev = $lru[$o['l']];
				$this->outlines[$prev]['next'] = $i;
				$this->outlines[$i]['prev'] = $prev;
			}
			$lru[$o['l']] = $i;
			$level = $o['l'];
		}

		// Outline items
		$n = $this->objectNumber + 1;
		foreach ($this->outlines as $i => $o) {
			$this->writeObject();
			$this->write('<</Title ' . $this->utf16BigEndianTextString($o['t']));
			$this->write('/Parent ' . ($n + $o['parent']) . ' 0 R');
			if (isset($o['prev'])) {
				$this->write('/Prev ' . ($n + $o['prev']) . ' 0 R');
			}
			if (isset($o['next'])) {
				$this->write('/Next ' . ($n + $o['next']) . ' 0 R');
			}
			if (isset($o['first'])) {
				$this->write('/First ' . ($n + $o['first']) . ' 0 R');
			}
			if (isset($o['last'])) {
				$this->write('/Last ' . ($n + $o['last']) . ' 0 R');
			}

			$anchor = str_replace("#", '\057', $o['d']);
			$this->write(sprintf('/Dest (%s)', $anchor));

			$this->write('/Count 0>>');
			$this->write('endobj');
		}
		// Outline root
		$this->writeObject();

		$this->write('<</Type /BMoutlines /First ' . $n . ' 0 R');
		$this->write('/Last ' . ($n + $lru[0]) . ' 0 R>>');
		$this->write('endobj');

		return $this->objectNumber;
	}

}